/*
*  ALCATEL SpeedTouch USB modem microcode upload & ADSL link UP utility
*  Copyright (C) 2001 Benoit PAPILLAULT
*  
*  This program is free software; you can redistribute it and/or
*  modify it under the terms of the GNU General Public License
*  as published by the Free Software Foundation; either version 2
*  of the License, or (at your option) any later version.
*  
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*  Author   : Benoit PAPILLAULT <benoit.papillault@free.fr>
*  Creation : 05/03/2001
*  
*  This program uploads the microcode to the ALCATEL SpeedTouch USB modem.
*  It uses libusb (http://libusb.sourceforge.net/).
*  
*  The microcode can be uploaded only once. If the upload is correct,
*  ADSL led should be both green & red.
*  
*  If you try to download the microcode twice, you will get tons of
*  timeout errors.
*
*  $Id: modem_run.c,v 1.31 2003/06/27 22:00:55 edgomez Exp $
*/

#ifndef _MODEM_RUN_C_
#define _MODEM_RUN_C_

#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>     /* for errno */
#include <stdarg.h>
#include <syslog.h>
#include <pwd.h>

#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "pusb.h"
#include "modem.h"
#include "pppoa3.h"
#include "extract.h"

/******************************************************************************
* Defines
******************************************************************************/

/* Timeout in milliseconds */
#define CTRL_TIMEOUT 2000
#define DATA_TIMEOUT 2000

#define OFFSET_7  0 /* size 1 */
#define OFFSET_b  1 /* size 8 */
#define OFFSET_d  9 /* size 4 */
#define OFFSET_e 13 /* size 1 */
#define OFFSET_f 14 /* size 1 */
#define TOTAL    15

#define SIZE_7 1
#define SIZE_b 8
#define SIZE_d 4
#define SIZE_e 1
#define SIZE_f 1

/* Remove the comments to enable old code parts */
/* #define PROPRIETARY_CODE */

/*****************************************************************************
* Global variables
******************************************************************************/

static int verbose = 0;
FILE *flog = NULL;

/*****************************************************************************
* Local Variables
*****************************************************************************/

static int dl_512_first = 1; /* Try to download 512 bytes before first op */
static int link_up = 0;      /* Used to gives link state between 2 processes */
static int sb = 0;           /* Software buffering */
static int signal_kernel = 0;/* Kernel driver notification */

/*
 * This array is just ARM code that initializes the speedtouch
 * and then waits for real firmware
 *
 * The first part is the proprietary code from Alcatel.
 * The second part is generated thanks to the Free Uploader code initmi.S
 */
static unsigned char data1[] =
{

#ifdef PROPRIETARY_CODE
	0x88, 0xf8, 0x00, 0x00, 0x00, 0x00, 0xf2, 0x01, 
	0x0d, 0x00, 0x00, 0xea, 0x0c, 0x00, 0x00, 0xea, 
	0x0b, 0x00, 0x00, 0xea, 0x0a, 0x00, 0x00, 0xea, 
	0x09, 0x00, 0x00, 0xea, 0x08, 0x00, 0x00, 0xea, 
	0xa4, 0x00, 0x00, 0xea, 0x06, 0x00, 0x00, 0xea, 
	0x32, 0x11, 0xa0, 0xe3, 0x5c, 0x03, 0x9f, 0xe5, 
	0x00, 0x00, 0x81, 0xe5, 0x58, 0x13, 0x9f, 0xe5, 
	0x00, 0x00, 0xa0, 0xe3, 0x00, 0x00, 0x81, 0xe5, 
	0xf8, 0xff, 0xff, 0xea, 0x01, 0x00, 0xa0, 0xe3, 
	0x48, 0x13, 0x9f, 0xe5, 0x00, 0x00, 0x81, 0xe5, 
	0x3c, 0x13, 0x9f, 0xe5, 0x04, 0x00, 0xa0, 0xe3, 
	0x00, 0x00, 0x81, 0xe5, 0xd1, 0x00, 0xa0, 0xe3, 
	0x00, 0xf0, 0x29, 0xe1, 0x01, 0xd9, 0xa0, 0xe3, 
	0xd2, 0x00, 0xa0, 0xe3, 0x00, 0xf0, 0x29, 0xe1, 
	0x24, 0xd3, 0x9f, 0xe5, 0xd3, 0x00, 0xa0, 0xe3, 
	0x00, 0xf0, 0x29, 0xe1, 0x1c, 0xd3, 0x9f, 0xe5, 
	0x00, 0x00, 0xa0, 0xe3, 0x00, 0x10, 0xa0, 0xe3, 
	0xb0, 0x02, 0x9f, 0xe5, 0x08, 0x00, 0x80, 0xe2, 
	0xbc, 0x12, 0x9f, 0xe5, 0x00, 0x10, 0x80, 0xe5, 
	0x5f, 0x10, 0xe0, 0xe3, 0x5d, 0x30, 0xe0, 0xe3, 
	0x5c, 0x50, 0xe0, 0xe3, 0x5b, 0x70, 0xe0, 0xe3, 
	0xac, 0x02, 0x9f, 0xe5, 0xa4, 0x22, 0x9f, 0xe5, 
	0xa8, 0x42, 0x9f, 0xe5, 0xa8, 0x62, 0x9f, 0xe5, 
	0x00, 0x10, 0x80, 0xe5, 0x00, 0x30, 0x82, 0xe5, 
	0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
	0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
	0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
	0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
	0x00, 0x70, 0x86, 0xe5, 0x54, 0x02, 0x9f, 0xe5, 
	0x08, 0x00, 0x80, 0xe2, 0x5c, 0x12, 0x9f, 0xe5, 
	0x00, 0x10, 0x80, 0xe5, 0x6c, 0x02, 0x9f, 0xe5, 
	0x6c, 0x12, 0x9f, 0xe5, 0x00, 0x10, 0x80, 0xe5, 
	0x00, 0x10, 0xa0, 0xe3, 0x98, 0x02, 0x9f, 0xe5, 
	0x00, 0x10, 0x80, 0xe5, 0x94, 0x02, 0x9f, 0xe5, 
	0x00, 0x10, 0x80, 0xe5, 0x01, 0x09, 0xa0, 0xe3, 
	0x00, 0x10, 0xa0, 0xe3, 0x02, 0x26, 0xa0, 0xe3, 
	0x04, 0x30, 0xa0, 0xe3, 0x03, 0x10, 0x80, 0xe6, 
	0x02, 0x00, 0x50, 0xe1, 0xfc, 0xff, 0xff, 0x1a, 
	0x7f, 0x19, 0xa0, 0xe3, 0x1e, 0x30, 0xa0, 0xe3, 
	0x03, 0x3b, 0x83, 0xe2, 0x00, 0x00, 0xa0, 0xe3, 
	0x00, 0x2a, 0x83, 0xe1, 0x01, 0x00, 0x80, 0xe2, 
	0x14, 0x00, 0x50, 0xe3, 0x04, 0x20, 0x81, 0xe4, 
	0xfa, 0xff, 0xff, 0xba, 0x02, 0xe0, 0xa0, 0xe3, 
	0x03, 0xeb, 0x8e, 0xe2, 0x0e, 0x20, 0xa0, 0xe1, 
	0x18, 0x00, 0x50, 0xe3, 0x04, 0x00, 0x00, 0xaa, 
	0x00, 0xca, 0x82, 0xe1, 0x01, 0x00, 0x80, 0xe2, 
	0x18, 0x00, 0x50, 0xe3, 0x04, 0xc0, 0x81, 0xe4, 
	0xfa, 0xff, 0xff, 0xba, 0x40, 0x00, 0x50, 0xe3, 
	0x04, 0x00, 0x00, 0xaa, 0x00, 0x2a, 0x83, 0xe1, 
	0x01, 0x00, 0x80, 0xe2, 0x40, 0x00, 0x50, 0xe3, 
	0x04, 0x20, 0x81, 0xe4, 0xfa, 0xff, 0xff, 0xba, 
	0x01, 0x0a, 0x50, 0xe3, 0x04, 0x00, 0x00, 0xaa, 
	0x00, 0x2a, 0x8e, 0xe1, 0x01, 0x00, 0x80, 0xe2, 
	0x01, 0x0a, 0x50, 0xe3, 0x04, 0x20, 0x81, 0xe4, 
	0xfa, 0xff, 0xff, 0xba, 0x7f, 0x09, 0xa0, 0xe3, 
	0x03, 0x40, 0xa0, 0xe3, 0x10, 0x0f, 0x02, 0xee, 
	0x10, 0x4f, 0x03, 0xee, 0x10, 0x0f, 0x07, 0xee, 
	0x10, 0x0f, 0x05, 0xee, 0x3d, 0x00, 0xa0, 0xe3, 
	0x10, 0x0f, 0x01, 0xee, 0x01, 0x73, 0xa0, 0xe3, 
	0x04, 0x00, 0xa0, 0xe3, 0x05, 0x00, 0xc7, 0xe5, 
	0x01, 0x00, 0xa0, 0xe3, 0x06, 0x00, 0xc7, 0xe5, 
	0x01, 0x00, 0xa0, 0xe3, 0x07, 0x00, 0xc7, 0xe5, 
	0x03, 0x00, 0xa0, 0xe3, 0x08, 0x00, 0xc7, 0xe5, 
	0x00, 0x00, 0x40, 0x01, 0x12, 0x88, 0xd4, 0xf2, 
	0x01, 0x00, 0x00, 0xce, 0x01, 0xa0, 0xe3, 0x09, 
	0x00, 0xc7, 0xe5, 0x80, 0x00, 0xa0, 0xe3, 0x01, 
	0x00, 0xc7, 0xe5, 0x53, 0x00, 0xa0, 0xe3, 0x00, 
	0xf0, 0x29, 0xe1, 0x94, 0x11, 0x9f, 0xe5, 0x21, 
	0x03, 0xa0, 0xe3, 0x00, 0x10, 0x80, 0xe5, 0x10, 
	0x10, 0xa0, 0xe3, 0x22, 0x01, 0xa0, 0xe3, 0x00, 
	0x10, 0x80, 0xe5, 0x04, 0x00, 0x8f, 0xe2, 0x01, 
	0x00, 0x80, 0xe2, 0x10, 0xff, 0x2f, 0xe1, 0x5e, 
	0x4f, 0xb8, 0x78, 0x40, 0x1e, 0xfc, 0xd1, 0x38, 
	0x78, 0xf9, 0x7a, 0x09, 0x02, 0xbc, 0x7a, 0x21, 
	0x43, 0x09, 0x02, 0x7c, 0x7a, 0x21, 0x43, 0x09, 
	0x02, 0x3c, 0x7a, 0x21, 0x43, 0x11, 0x38, 0x01, 
	0x28, 0x2d, 0xdb, 0x03, 0x28, 0x2b, 0xdc, 0x80, 
	0x00, 0x55, 0x4c, 0x24, 0x58, 0xa7, 0x46, 0x46, 
	0x4b, 0x0e, 0x33, 0x7a, 0x7b, 0x12, 0x02, 0x3c, 
	0x7b, 0x22, 0x43, 0x12, 0xd0, 0x14, 0x1c, 0xf4, 
	0x3c, 0xff, 0x3c, 0x0e, 0xd5, 0x40, 0x07, 0x02, 
	0xd0, 0x1c, 0x1c, 0x0b, 0x1c, 0x21, 0x1c, 0x52, 
	0x1e, 0x8c, 0x5c, 0x9c, 0x54, 0xfb, 0xd1, 0x3b, 
	0xa4, 0x14, 0x23, 0x23, 0x60, 0xc0, 0x0f, 0x64, 
	0x30, 0x00, 0xe0, 0x66, 0x20, 0x78, 0x70, 0xcb, 
	0xe7, 0x01, 0xa2, 0x10, 0x47, 0x00, 0x00, 0x01, 
	0xf0, 0xa0, 0xe1, 0x04, 0x20, 0x8f, 0xe2, 0x01, 
	0x20, 0x82, 0xe2, 0x12, 0xff, 0x2f, 0xe1, 0x60, 
	0x20, 0xf0, 0xe7, 0x04, 0xe0, 0x4e, 0xe2, 0x00, 
	0x40, 0x2d, 0xe9, 0x00, 0xe0, 0x4f, 0xe1, 0x07, 
	0x40, 0x2d, 0xe9, 0xa8, 0x10, 0x8f, 0xe2, 0x00, 
	0x00, 0x91, 0xe5, 0x01, 0x00, 0x20, 0xe2, 0x00, 
	0x00, 0x81, 0xe5, 0x00, 0x00, 0x50, 0xe3, 0x03, 
	0x00, 0x00, 0x0a, 0x0d, 0x10, 0xa0, 0xe3, 0xcc, 
	0x20, 0x9f, 0xe5, 0x00, 0x10, 0x82, 0xe5, 0x02, 
	0x00, 0x00, 0xea, 0x0f, 0x10, 0xa0, 0xe3, 0xbc, 
	0x20, 0x9f, 0xe5, 0x00, 0x10, 0x82, 0xe5, 0x01, 
	0x00, 0xa0, 0xe3, 0x90, 0x10, 0x9f, 0xe5, 0x00, 
	0x00, 0x81, 0xe5, 0x6c, 0x10, 0x8f, 0xe2, 0x00, 
	0x00, 0x91, 0xe5, 0x01, 0x00, 0x40, 0xe2, 0x00, 
	0x00, 0x81, 0xe5, 0x00, 0x00, 0x50, 0xe3, 0x02, 
	0x00, 0x00, 0x1a, 0x94, 0x00, 0x9f, 0xe5, 0x94, 
	0x10, 0x9f, 0xe5, 0x00, 0x00, 0x81, 0xe5, 0x90, 
	0x20, 0x9f, 0xe5, 0x00, 0x10, 0x92, 0xe5, 0x07, 
	0x40, 0xbd, 0xe8, 0x0e, 0xf0, 0x69, 0xe1, 0x00, 
	0x80, 0xfd, 0xe8, 0x00, 0x00, 0x00, 0x8c, 0x20, 
	0x00, 0x00, 0x88, 0x50, 0x00, 0x00, 0x01, 0x02, 
	0x00, 0x00, 0x00, 0xf4, 0x00, 0x00, 0x00, 0x39, 
	0x00, 0x00, 0x00, 0x10, 0x40, 0x08, 0x00, 0x00, 
	0x40, 0x00, 0x00, 0x20, 0x40, 0x00, 0x00, 0x30, 
	0x4e, 0x00, 0x00, 0x00, 0x00, 0x00, 0xa0, 0x15, 
	0x80, 0x00, 0x00, 0x15, 0x2f, 0x8c, 0x9b, 0x00, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x78, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x5d, 
	0x02, 0x00, 0x00, 0x5d, 0x02, 0x00, 0x00, 0x97, 
	0x02, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0x40, 
	0x00, 0x40, 0x01, 0x64, 0x00, 0x50, 0x01, 0x70, 
	0x3e, 0x00, 0x00, 0xe0, 0x3c, 0x00, 0x00, 0x20, 
	0x00, 0x00, 0x90, 0x20, 0x00, 0x00, 0xa4, 0x85, 
	0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x78, 
	0x03, 0x00, 0x00, 0x58, 0x00, 0x50, 0x01, 0x70, 
	0xc3, 0x00, 0x00, 0x60, 0x00, 0x50, 0x01, 0x04, 
	0x00, 0x00, 0x84, 0x40, 0x01, 0x12, 0x08, 0x04, 
	0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x14
#else
	0x88, 0xf8, 0x00, 0x00, 0x00, 0x00, 0xf2, 0x01, 
	0x06, 0x00, 0x00, 0xea, 0x05, 0x00, 0x00, 0xea, 
	0x04, 0x00, 0x00, 0xea, 0x03, 0x00, 0x00, 0xea, 
	0x02, 0x00, 0x00, 0xea, 0x01, 0x00, 0x00, 0xea, 
	0xa9, 0x00, 0x00, 0xea, 0xff, 0xff, 0xff, 0xea, 
	0x01, 0x00, 0xa0, 0xe3, 0x15, 0x16, 0xa0, 0xe3, 
	0x64, 0x10, 0x81, 0xe3, 0x00, 0x00, 0x81, 0xe5, 
	0x04, 0x00, 0xa0, 0xe3, 0x05, 0x15, 0xa0, 0xe3, 
	0x40, 0x10, 0x81, 0xe3, 0x00, 0x00, 0x81, 0xe5, 
	0xd1, 0x00, 0xa0, 0xe3, 0x00, 0xf0, 0x29, 0xe1, 
	0x01, 0xd9, 0xa0, 0xe3, 0xd2, 0x00, 0xa0, 0xe3, 
	0x00, 0xf0, 0x29, 0xe1, 0x3e, 0xdc, 0xa0, 0xe3, 
	0x70, 0xd0, 0x8d, 0xe3, 0xd3, 0x00, 0xa0, 0xe3, 
	0x00, 0xf0, 0x29, 0xe1, 0x0f, 0xdb, 0xa0, 0xe3, 
	0xe0, 0xd0, 0x8d, 0xe3, 0x23, 0x03, 0xa0, 0xe3, 
	0x08, 0x00, 0x80, 0xe2, 0x39, 0x10, 0xa0, 0xe3, 
	0x00, 0x10, 0x80, 0xe5, 0x5f, 0x10, 0xe0, 0xe3, 
	0x5d, 0x30, 0xe0, 0xe3, 0x5c, 0x50, 0xe0, 0xe3, 
	0x5b, 0x70, 0xe0, 0xe3, 0x21, 0x29, 0xa0, 0xe3, 
	0x10, 0x20, 0x82, 0xe3, 0x01, 0x09, 0xa0, 0xe3, 
	0x20, 0x40, 0x80, 0xe3, 0xe3, 0x6e, 0x80, 0xe3, 
	0x00, 0x10, 0x80, 0xe5, 0x00, 0x30, 0x82, 0xe5, 
	0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
	0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
	0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
	0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
	0x00, 0x70, 0x86, 0xe5, 0x23, 0x03, 0xa0, 0xe3, 
	0x08, 0x00, 0x80, 0xe2, 0xf4, 0x10, 0xa0, 0xe3, 
	0x00, 0x10, 0x80, 0xe5, 0x0a, 0x02, 0xa0, 0xe3, 
	0x02, 0x19, 0xa0, 0xe3, 0x15, 0x10, 0x81, 0xe3, 
	0x00, 0x10, 0x80, 0xe5, 0x01, 0x10, 0xa0, 0xe3, 
	0x09, 0x02, 0xa0, 0xe3, 0x20, 0x00, 0x80, 0xe3, 
	0x00, 0x10, 0x80, 0xe5, 0x29, 0x03, 0xa0, 0xe3, 
	0x20, 0x00, 0x80, 0xe3, 0x00, 0x10, 0x80, 0xe5, 
	0x02, 0x06, 0xa0, 0xe3, 0x00, 0x10, 0xa0, 0xe3, 
	0x01, 0x09, 0x50, 0xe3, 0x04, 0x10, 0x00, 0xe4, 
	0xfc, 0xff, 0xff, 0xca, 0x00, 0x00, 0xa0, 0xe3, 
	0x7f, 0x19, 0xa0, 0xe3, 0xc1, 0x3e, 0xa0, 0xe3, 
	0x0e, 0x30, 0x83, 0xe3, 0x00, 0x2a, 0x83, 0xe1, 
	0x01, 0x00, 0x80, 0xe2, 0x04, 0x20, 0x81, 0xe4, 
	0x14, 0x00, 0x50, 0xe3, 0xfa, 0xff, 0xff, 0xba, 
	0x03, 0x3b, 0xa0, 0xe3, 0x02, 0x30, 0x83, 0xe3, 
	0x00, 0x2a, 0x83, 0xe1, 0x01, 0x00, 0x80, 0xe2, 
	0x04, 0x20, 0x81, 0xe4, 0x18, 0x00, 0x50, 0xe3, 
	0xfa, 0xff, 0xff, 0xba, 0xc1, 0x3e, 0xa0, 0xe3, 
	0x0e, 0x30, 0x83, 0xe3, 0x00, 0x2a, 0x83, 0xe1, 
	0x01, 0x00, 0x80, 0xe2, 0x04, 0x20, 0x81, 0xe4, 
	0x40, 0x00, 0x50, 0xe3, 0xfa, 0xff, 0xff, 0xba, 
	0x03, 0x3b, 0xa0, 0xe3, 0x02, 0x30, 0x83, 0xe3, 
	0x00, 0x2a, 0x83, 0xe1, 0x01, 0x00, 0x80, 0xe2, 
	0x04, 0x20, 0x81, 0xe4, 0x01, 0x0a, 0x50, 0xe3, 
	0xfa, 0xff, 0xff, 0xba, 0x7f, 0x09, 0xa0, 0xe3, 
	0x03, 0x40, 0xa0, 0xe3, 0x10, 0x0f, 0x02, 0xee, 
	0x10, 0x4f, 0x03, 0xee, 0x10, 0x0f, 0x07, 0xee, 
	0x10, 0x0f, 0x05, 0xee, 0x3d, 0x00, 0xa0, 0xe3, 
	0x10, 0x0f, 0x01, 0xee, 0x01, 0x73, 0xa0, 0xe3, 
	0x04, 0x00, 0xa0, 0xe3, 0x05, 0x00, 0xc7, 0xe5, 
	0x01, 0x00, 0xa0, 0xe3, 0x06, 0x00, 0xc7, 0xe5, 
	0x07, 0x00, 0xc7, 0xe5, 0x03, 0x00, 0xa0, 0xe3, 
	0x08, 0x00, 0xc7, 0xe5, 0x00, 0x00, 0xa0, 0xe3, 
	0x09, 0x00, 0xc7, 0xe5, 0x80, 0x00, 0xa0, 0xe3, 
	0x01, 0x00, 0xc7, 0xe5, 0x53, 0x00, 0xa0, 0xe3, 
	0x00, 0xf0, 0x29, 0xe1, 0x01, 0x19, 0xa0, 0xe3, 
	0x85, 0x10, 0x40, 0x01, 0x12, 0x88, 0x68, 0xf2, 
	0x01, 0x00, 0x00, 0x62, 0x01, 0x81, 0xe3, 0x21, 
	0x03, 0xa0, 0xe3, 0x00, 0x10, 0x80, 0xe5, 0x10, 
	0x10, 0xa0, 0xe3, 0x22, 0x01, 0xa0, 0xe3, 0x00, 
	0x10, 0x80, 0xe5, 0x02, 0x00, 0xd7, 0xe5, 0x01, 
	0x00, 0x50, 0xe2, 0xfc, 0xff, 0xff, 0x1a, 0x00, 
	0x00, 0xd7, 0xe5, 0x11, 0x00, 0x40, 0xe2, 0x0b, 
	0x10, 0xd7, 0xe5, 0x0a, 0x40, 0xd7, 0xe5, 0x01, 
	0x14, 0x84, 0xe1, 0x09, 0x40, 0xd7, 0xe5, 0x01, 
	0x14, 0x84, 0xe1, 0x08, 0x40, 0xd7, 0xe5, 0x01, 
	0x14, 0x84, 0xe1, 0x01, 0x00, 0x50, 0xe3, 0x60, 
	0x00, 0xa0, 0xb3, 0x1d, 0x00, 0x00, 0xba, 0x03, 
	0x00, 0x50, 0xe3, 0x60, 0x00, 0xa0, 0xc3, 0x1a, 
	0x00, 0x00, 0xca, 0x03, 0x00, 0x50, 0xe3, 0x01, 
	0xf0, 0xa0, 0x01, 0x0d, 0x20, 0xd7, 0xe5, 0x0c, 
	0x40, 0xd7, 0xe5, 0x02, 0x24, 0x94, 0xe1, 0x13, 
	0x00, 0x00, 0x0a, 0x1f, 0x4e, 0xa0, 0xe3, 0x02, 
	0x40, 0x84, 0xe3, 0x04, 0x00, 0x52, 0xe1, 0x0f, 
	0x00, 0x00, 0xca, 0x07, 0x30, 0xa0, 0xe1, 0x0e, 
	0x30, 0x83, 0xe2, 0x01, 0x00, 0x50, 0xe3, 0x02, 
	0x00, 0x00, 0x1a, 0x03, 0x40, 0xa0, 0xe1, 0x01, 
	0x30, 0xa0, 0xe1, 0x04, 0x10, 0xa0, 0xe1, 0x01, 
	0x20, 0x52, 0xe2, 0x02, 0x40, 0xd1, 0xe7, 0x02, 
	0x40, 0xc3, 0xe7, 0xfb, 0xff, 0xff, 0x1a, 0x14, 
	0x30, 0xa0, 0xe3, 0x9c, 0x30, 0x8f, 0xe5, 0x01, 
	0x00, 0x00, 0xe2, 0x64, 0x00, 0x80, 0xe2, 0x00, 
	0x00, 0x00, 0xea, 0x66, 0x00, 0xa0, 0xe3, 0x01, 
	0x00, 0xc7, 0xe5, 0xd0, 0xff, 0xff, 0xea, 0x04, 
	0xe0, 0x4e, 0xe2, 0x00, 0x40, 0x2d, 0xe9, 0x00, 
	0xe0, 0x4f, 0xe1, 0x07, 0x40, 0x2d, 0xe9, 0x6c, 
	0x00, 0x9f, 0xe5, 0x01, 0x00, 0x20, 0xe2, 0x64, 
	0x00, 0x8f, 0xe5, 0x00, 0x00, 0x50, 0xe3, 0x64, 
	0x10, 0x9f, 0x05, 0x0f, 0x10, 0xa0, 0x13, 0x15, 
	0x26, 0xa0, 0xe3, 0x58, 0x20, 0x82, 0xe3, 0x00, 
	0x10, 0x82, 0xe5, 0x01, 0x00, 0xa0, 0xe3, 0x15, 
	0x16, 0xa0, 0xe3, 0x64, 0x10, 0x81, 0xe3, 0x00, 
	0x00, 0x81, 0xe5, 0x3c, 0x00, 0x9f, 0xe5, 0x01, 
	0x00, 0x40, 0xe2, 0x34, 0x00, 0x8f, 0xe5, 0x00, 
	0x00, 0x50, 0xe3, 0x04, 0x00, 0x00, 0x1a, 0xc3, 
	0x0c, 0xa0, 0xe3, 0x70, 0x00, 0x80, 0xe3, 0x15, 
	0x16, 0xa0, 0xe3, 0x60, 0x10, 0x81, 0xe3, 0x00, 
	0x00, 0x81, 0xe5, 0x21, 0x23, 0xa0, 0xe3, 0x04, 
	0x20, 0x82, 0xe3, 0x00, 0x10, 0x92, 0xe5, 0x07, 
	0x40, 0xbd, 0xe8, 0x0e, 0xf0, 0x69, 0xe1, 0x00, 
	0x80, 0xfd, 0xe8, 0x00, 0x00, 0x00, 0x00, 0x78, 
	0x00, 0x00, 0x00, 0x0d, 0x00, 0x00, 0x00, 0x40, 
	0x01, 0x12, 0x08, 0x04, 0x00, 0x00, 0x00, 0x00, 
	0x00, 0x01, 0x14
#endif
};

/******************************************************************************
* Prototypes
******************************************************************************/

static int  upload_microcode(pusb_device_t fdusb,const char * filename);
static int  get_reference(pusb_device_t fdusb);
static int  modem_start_synchro(pusb_device_t fdusb);
static void swbuff(pusb_device_t fdusb, int state);
static int  get_state(pusb_device_t fdusb, unsigned char *buf);
static int  print_state(unsigned char *buf);

static void usage();
static void test_sequence(pusb_device_t fdusb);
static void handle_endpoint_int(pusb_device_t fdusb);


static void report_stop();
static void dump(unsigned char *buf, int len);
static unsigned char give_char(unsigned char c);

static void signal_usr2(int signal);

/*****************************************************************************
*	Main function
*****************************************************************************/

/*
* Function      : main
* Return Values : ?
* Description   : Does all the job (that's fantastic :)
*/
int main(int argc, char *argv[])
{

	pusb_device_t fdusb;
	int i;
	int monitoring = 0;
	unsigned char buf[TOTAL+1000];
	const char * file = NULL;
	char *devicename = NULL;
	char *user;
	
	/*
	* Security stuff
	* 1 - be sure to be root
	* 2 - umask to prevent critical data being read from log file
	*/
	if(geteuid() != 0) {
		fprintf(stderr, "modem_run must be run with root privileges\n");
		return(-1);
	}

	/* Gets user login */
	user = getlogin();

	if(user == NULL) {

		struct passwd *pw;

		pw = getpwuid(getuid());

		if( pw != NULL && pw->pw_name != NULL)
			user = pw->pw_name;
		else
			user = "Unknown";
	}

	report(0, REPORT_INFO, "modem_run version %s started by %s uid %d\n", VERSION, user, getuid());

	for (i=1;i<argc;i++) {

		if (strcmp(argv[i],"-v") == 0 && i+1<argc)
			verbose = atoi(argv[++i]);
		else if (strcmp(argv[i],"-b")==0)
			sb = 1;
		else if (strcmp(argv[i],"-k")==0)
			signal_kernel = 1;
		else if (strcmp(argv[i],"-m")==0)
			monitoring = 1;
		else if (strcmp(argv[i],"-s")==0)
			dl_512_first = 0;
		else if (strcmp(argv[i],"-f")==0 && i+1<argc)
			file = argv[++i];
		else if(strcmp(argv[i], "-d") == 0 && i + 1 < argc)
			devicename = argv[++i];
		else if (strcmp(argv[i],"--help")==0)
			usage();
		else
			usage();

	}
	
	if (file == NULL)
		usage();

	/* we search the modem on the USB bus */
	if (devicename == NULL) {

		fdusb = pusb_search_open(ST_VENDOR,ST_PRODUCT);

		if (fdusb == NULL) {
			report(0, REPORT_ERROR, "No SpeedTouch USB found.\n");
			return(-1);
		}

	}
	else {

		fdusb = pusb_open(devicename);

		if(fdusb == NULL) {
			report(0, REPORT_ERROR, "No SpeedTouch USB at %s\n", devicename);
			return(-1);
		}

	}
	
	report(1, REPORT_INFO, "Found ALCATEL SpeedTouch USB modem\n");

	/*
	 * we check that no one else is already using the modem,
	 * by claiming (ie requesting exclusive use) interface 0, 1 & 2,
	 * which are all the interfaces of the USB modem.
	 */
	
	if (pusb_claim_interface(fdusb,2) < 0) {
		report(0, REPORT_ERROR, "Another program/driver is already accessing the modem (interface 2 cannot be claimed)...\n");
		return(-1);
	}

	if (pusb_set_configuration(fdusb,1) < 0) {
		report(0, REPORT_ERROR, "pusb_set_configuration 1");
		return(-1);
	}
	

	/*
	 * For some extremely rare modems, the first upload fails
	 * The workaround is simple : we try 10 times (10 is magic)
	 */
	for(i=0; i<10; i++) {

		if (upload_microcode (fdusb,file) < 0) {
			report(1, REPORT_ERROR, "Retrying to upload microcode (#%d)\n", i);
			continue;
		}

		break;
	}

	/* It failed all the times */
	if(i == 10) {
			
		if (pusb_release_interface(fdusb,2) < 0)
			perror("pusb_release_interface 2");
	
		pusb_close(fdusb);
		report_stop();
		return(-1);

	}
	
	/*
	 * Ok the microcode is loaded, wait a bit before asking him
	 * the adsl sync
	 */
	sleep(1);
	
	if (pusb_claim_interface(fdusb,0) < 0) {
		report(0, REPORT_ERROR, "Another program/driver is already accessing the modem (interface 0 cannot be claimed)...\n");
		return(-1);
	}

	/* reconfigure USB (configuration & alternate settings) */
	if (pusb_set_configuration(fdusb, 1) < 0) {
		report(0, REPORT_ERROR, "pusb_set_configuration 1");
		return(-1);
	}
	
	if (pusb_claim_interface(fdusb,1) < 0) {
		if (!signal_kernel) {
			report(0, REPORT_ERROR, "Another program/driver is already accessing the modem (interface 1 cannot be claimed)...\n");
			return(-1);
		}
		report(1, REPORT_INFO, "Found kernel mode driver\n");
	} else { /* claimed interface 1 */
		if (signal_kernel) {
			report(0, REPORT_ERROR, "Kernel mode driver not found\n"); 
			signal_kernel = 0;
		}

		/*
		 * We use alternate 1 so both "classic" and new SpeedTouch USB 330
		 * can be initialised.
		 *
		 * NB: Historically we used alternate 2 (before 2002-12-13)
		 */
		if (pusb_set_interface(fdusb, 1, 1) < 0) {
			report(0, REPORT_ERROR, "pusb_set_interface");
			return(-1);
		}
	}

	/* Enable software buffering - Courtesy of Thomson Multimedia */
	if(sb) swbuff(fdusb,1);

	/* 
	 * We register the signal handler that will be used to signal
	 * the up state of the link.
	 */
	signal(SIGUSR2 , signal_usr2);

	/*
	 * Start a subprocess (the so called daemon) to make sure the
	 * interrupt endpoint is always writable (because we read it there)
	 */
	handle_endpoint_int(fdusb);

	/* Cosmetic */
	get_reference(fdusb);
	test_sequence(fdusb);
	
	/* should start line sync */
	modem_start_synchro(fdusb);

	/*
	 * New method : The child which reads the interrupt endpoint
	 *              notices the parent process of the up link
	 */
	while(!link_up)
		pause();

	/* The physical adsl link has been established */
	report(0, REPORT_INFO, "ADSL synchronization has been obtained\n");

	/* We can print the adsl link state (speed) */
	get_state(fdusb,buf);
	print_state(buf);

	/* Close all */
	if (pusb_release_interface(fdusb,0) < 0)
		report(0, REPORT_ERROR, "pusb_release_interface 0 failed");

	if (!signal_kernel && pusb_release_interface(fdusb,1) < 0)
		report(0, REPORT_ERROR, "pusb_release_interface 1 failed");

	if (pusb_release_interface(fdusb,2) < 0)
		report(0, REPORT_ERROR, "pusb_release_interface 2 failed");
	
	pusb_close(fdusb);

	report_stop();
	
	return(0);

}

/*****************************************************************************
*	Subroutines
*****************************************************************************/

/*
* Function      : usage
* Return Values : None
* Description   : Prints an usage message
*/
static void usage()
{

	fprintf(stderr, "modem_run version %s:  $\n", VERSION);
	fprintf(stderr, "usage: modem_run [-v level] [-m] -f microcode \n");
	fprintf(stderr, "  -v           : verbose level [0-2]\n");
	fprintf(stderr, "  -m           : monitor the line\n");
	fprintf(stderr, "  -s           : skip the first 512 bytes read\n");
	fprintf(stderr, "  -f microcode : upload this microcode file first\n");
	fprintf(stderr, "  -d device    : defines the device to use\n");
	fprintf(stderr, "  -k           : using the kernel mode driver\n");
	fprintf(stderr, "  --help       : this message\n");
	exit(-1);

}

/*
* Function      : upload_microcode
* Return Values : -1 in case of error
* Description   : [...]
* NB            : This functions require the modem interface 2
*/
static int upload_microcode(pusb_device_t fdusb,const char * filename)
{


	int uc_len;
	int n;
	unsigned char *uc_buf, buf[0x1000];
	/* buf should be at least 511 bytes */
	struct timeval start, stop;
	pusb_endpoint_t ep_code;

	/* Retrieves the microcode from file "filename" */
	if ((uc_buf = extract_microcode(filename,&uc_len)) == NULL) {

		/*
		* this is a fatal error, either the file cannot be read
		* or the microcode is not inside
		*/
		report(0, REPORT_ERROR, "Unable to locate microcode\n");
		exit(-1);
		
	}

	/* open the EP_CODE endpoint */
	if ((ep_code = pusb_endpoint_open(fdusb,EP_CODE_IN,O_RDWR)) == NULL) {
		report(0, REPORT_ERROR, "pusb_endpoint_open");
		exit(-1);
	}
	
	/* URB 7 */
	if(dl_512_first){

		memset(buf, 0, sizeof(buf));
		gettimeofday(&start,NULL);
		n = pusb_endpoint_read(ep_code,buf,0x200,DATA_TIMEOUT);
		gettimeofday(&stop,NULL);

		if (n < 0) {

			if (errno != ETIMEDOUT) {
				report(0, REPORT_ERROR, "BLOCK0\n");
				pusb_endpoint_close(ep_code);
				return(n);
			}

		}
		else {

			report(1, REPORT_INFO, "BLOCK0 : %6d bytes downloaded : OK\n", n);
			report(2, REPORT_DEBUG|REPORT_DUMP,"Downloaded in %f ms\n",
			       buf,
			       n,
			       (double)(stop.tv_sec -start.tv_sec) * 1000.0 +
			       (double)(stop.tv_usec - start.tv_usec) / 1000.0);

		}

	}
	
	/* URB 8 : both leds are static green */
	gettimeofday(&start,NULL);
	n = pusb_endpoint_write(ep_code,data1,sizeof(data1),DATA_TIMEOUT);
	gettimeofday(&stop,NULL);

	if (n < 0) {
		report(0, REPORT_ERROR, "BLOCK1\n");
		pusb_endpoint_close(ep_code);
		return(n);
	}
	else {
		report(1, REPORT_INFO, "BLOCK1 : %6d bytes   uploaded : OK\n", sizeof(data1));
		report(2, REPORT_DEBUG|REPORT_DUMP,"Downloaded in %f ms\n",
		       data1,
		       sizeof(data1),
		       (double)(stop.tv_sec -start.tv_sec) * 1000.0 +
		       (double)(stop.tv_usec - start.tv_usec) / 1000.0);

	}
	
	/*
	* After usb_bulk_write :
	*   + USB  led : blinking green
	*   + ADSL led : off
	*/
	
	/* URB 11 */
	memset(buf, 0, sizeof(buf));
	gettimeofday(&start,NULL);
	n = pusb_endpoint_read(ep_code,buf,0x200,DATA_TIMEOUT);
	gettimeofday(&stop,NULL);

	if (n < 0) {
		report(0, REPORT_ERROR, "BLOCK2\n");
		pusb_endpoint_close(ep_code);
		return(n);
	}
	else {

		report(1, REPORT_INFO, "BLOCK2 : %6d bytes downloaded : OK\n", n);
		report(2, REPORT_DEBUG|REPORT_DUMP,"Downloaded in %f ms\n",
		       buf,
		       n,
		       (double)(stop.tv_sec -start.tv_sec) * 1000.0 +
		       (double)(stop.tv_usec - start.tv_usec) / 1000.0);

	}
	
	
	/* URB 12 to 139 */
	gettimeofday(&start,NULL);
	n = pusb_endpoint_write(ep_code,uc_buf,uc_len,DATA_TIMEOUT);
	gettimeofday(&stop,NULL);
	
	if (n < 0) {
		report(0, REPORT_ERROR, "BLOCK3\n");
		pusb_endpoint_close(ep_code);
		return(n);
	}
	else {

		report(1, REPORT_INFO, "BLOCK3 : %6d bytes   uploaded : OK\n", n);
		report(2, REPORT_DEBUG, "Downloaded in %f ms\n", 
		       (double)(stop.tv_sec -start.tv_sec) * 1000.0 +
		       (double)(stop.tv_usec - start.tv_usec) / 1000.0);

	}

	/* Free the firmware allocated mem */
	free(uc_buf);

	/*
	* Before downloading microcode :
	*   + USB  led : blinking
	*   + ADSL led : off
	*   
	* During downloading:
	*   + USB  led : blinking green
	*   + ADSL led : off
	*   
	* After downloading microcode :
	*   + USB  led : static green
	*   + ADSL led : static red
	*/
	
	/* URB 142 */
	memset(buf, 0, sizeof(buf));
	gettimeofday(&start,NULL);
	n = pusb_endpoint_read(ep_code,buf,0x200,DATA_TIMEOUT);
	gettimeofday(&stop,NULL);
	
	if (n < 0) {
		report(0, REPORT_ERROR, "BLOCK4\n");
		pusb_endpoint_close(ep_code);
		return n;
	}
	else {

		report(1, REPORT_INFO, "BLOCK4 : %6d bytes downloaded : OK", n);
		report(2, REPORT_DEBUG|REPORT_DUMP, "Downloaded in %f ms.\n",
		       buf,
		       n,
		       (double)(stop.tv_sec -start.tv_sec) * 1000.0 +
		       (double)(stop.tv_usec - start.tv_usec) / 1000.0);

	}

	pusb_endpoint_close(ep_code);

	return(0);

}

/*
* Function      : get_reference
* Return Values : <0 on error, >0 on success
* Description   : Retrieves modem serial number
*/
static int get_reference(pusb_device_t fdusb)
{

	char buf[0xe];
	int i,n;
	
	/*
	* URB 151 : this buffer match (if you reverse one byte with its neighbour)
	* the label on the manta box!!! in my case : 3EC18607CAAB06
	* 3EC 18607CAAB 09
	* NB : "My case " == "Benoit's case"
	*/
	
	/* original line */
	n = pusb_control_msg(fdusb,0xc0,0x26,0xf7,0,buf,sizeof(buf),CTRL_TIMEOUT);
	if (n<0)
		report(0, REPORT_ERROR, "get_reference: pusb_control_msg\n");
	else {
	
		char temp[128];

		for (i=0;i<n;i++) {
			int idx = ((i%2)==0) ? i+1 : i-1;
			temp[i] = give_char(buf[idx]);
		}

		temp[n] = '\0';

		report(1, REPORT_INFO, "Modem reference : %s\n", temp);

	}

	return(n);

}

/*
* Function      : modem_start_synchro
* Return Values : -1 on error, 0 on success
* Description   : This function initiates the modem synchronisation
*/
static int modem_start_synchro(pusb_device_t fdusb)
{

	unsigned char buf[2];
	int n;

	n = pusb_control_msg(fdusb,0xc0,0x12,0x04,0x00,buf,sizeof(buf),CTRL_TIMEOUT);

	if (n < 0) {
		report(0, REPORT_ERROR, "modem_start_synchro: pusb_control_msg");
		return(-1);
	}

	report(2, REPORT_DEBUG|REPORT_DUMP, "modem_start_synchro :",
	       buf,
	       n);

	return(0);

}


/*
* Function      : swbuff ON/OFF
* Return Values : None
* Description   : Enables/disables software buffering
*                 Courtesy of Thomson Multimedia
*/  
static void swbuff(pusb_device_t fdusb, int state)
{
   int n;

   n = pusb_control_msg(fdusb,0x40,0x32,state?0x01:0x00,0x00,NULL,0,100);

   report(0, n<0 ? REPORT_ERROR : REPORT_INFO, 
          state ? "Enabling SW buffering\n" : "Disabling SW buffering\n");
 }

/*
* Function      : get_state
* Return Values : -1 is returned on errors, 0 otherwise
* Description   : Retrieves modem state from the usb interrupt end point
* NB            : buf MUST BE at least 15 bytes.
*/
static int get_state(pusb_device_t fdusb, unsigned char *buf)
{

	int ret;

	memset(buf,0,TOTAL);

	ret=pusb_control_msg(fdusb,0xc0,0x12,0x07,0x00,buf+OFFSET_7,SIZE_7,CTRL_TIMEOUT);
	if (ret<0) return(ret);

	ret=pusb_control_msg(fdusb,0xc0,0x12,0x0b,0x00,buf+OFFSET_b,SIZE_b,CTRL_TIMEOUT);
	if (ret<0) return(ret);

	ret=pusb_control_msg(fdusb,0xc0,0x12,0x0d,0x00,buf+OFFSET_d,SIZE_d,CTRL_TIMEOUT);
	if (ret<0) return(ret);

	ret=pusb_control_msg(fdusb,0xc0,0x01,0x0e,0x00,buf+OFFSET_e,SIZE_e,CTRL_TIMEOUT);
	if (ret<0) return(ret);

	ret=pusb_control_msg(fdusb,0xc0,0x01,0x0f,0x00,buf+OFFSET_f,SIZE_f,CTRL_TIMEOUT);
	if (ret<0) return(ret);

	return(0);

}

/*
* Function      : test_sequence
* Return Values : None
* Description   : This a magic sequence, totaly unknown to me [benoit],
*                 it comes from the Windows driver ... 
*/
static void test_sequence(pusb_device_t fdusb)
{

	unsigned char buf[10];
	int n;

	/* URB 147 */
	buf[0] = 0x1c; buf[1] = 0x50;
	n = pusb_control_msg(fdusb,0x40,0x01,0x0b,0x00,buf,2,100);
	if (n<0)
		report(0, REPORT_ERROR, "URB147\n");

	/* URB 148 */
	buf[0] = 0x32; buf[1] = 0x00;
	n = pusb_control_msg(fdusb,0x40,0x01,0x02,0x00,buf,2,100);
	if (n<0)
		report(0, REPORT_ERROR,"URB148\n");

	/* URB 149 */
	buf[0] = 0x01; buf[1] = 0x00; buf[2] = 0x01;
	n = pusb_control_msg(fdusb,0x40,0x01,0x03,0x00,buf,3,100);
	if (n<0)
		report(0, REPORT_ERROR, "URB149\n");
	
	/* URB 150 */
	buf[0] = 0x01; buf[1] = 0x00; buf[2] = 0x01;
	n = pusb_control_msg(fdusb,0x40,0x01,0x04,0x00,buf,3,100);
	if (n<0)
		report(0, REPORT_ERROR, "URB150\n");

}

/*
* Function      : print_state
* Return Values : 0
* Description   : Print a readale modem state
* NB            : buf MUST be 15 bytes long
*/
static int print_state(unsigned char *buf)
{
	time_t t = time(NULL);
	int down_speed = 0, up_speed = 0;
	
	report(2, REPORT_DEBUG,"Modem State\n",ctime(&t));
	report(2, REPORT_DEBUG|REPORT_DUMP,"c0 12 0007 : ", buf+OFFSET_7,SIZE_7);
	report(2, REPORT_DEBUG|REPORT_DUMP,"c0 12 000b : ", buf+OFFSET_b,SIZE_b);
	report(2, REPORT_DEBUG|REPORT_DUMP,"c0 12 000d : ", buf+OFFSET_d,SIZE_d);
	report(2, REPORT_DEBUG|REPORT_DUMP,"c0 01 000e : ", buf+OFFSET_e,SIZE_e);
	report(2, REPORT_DEBUG|REPORT_DUMP,"c0 01 000f : ", buf+OFFSET_f,SIZE_f);
	
	switch (buf[OFFSET_7]) {
	case 0x00:
		report(1, REPORT_INFO, "ADSL line is down\n");
		break;
	case 0x08:
		report(1, REPORT_INFO, "ADSL line is blocked?\n");
		break;
	case 0x10:
		report(1, REPORT_INFO, "ADSL line is synchronising\n");
		break;
	case 0x20:
		down_speed = buf[OFFSET_b] | (buf[OFFSET_b+1]<<8)
			| (buf[OFFSET_b+2]<<16) | (buf[OFFSET_b+3]<<24);
		up_speed = buf[OFFSET_b+4] | (buf[OFFSET_b+5]<<8)
			| (buf[OFFSET_b+6]<<16) | (buf[OFFSET_b+7]<<24);
		
	        if(	((down_speed & 0x0000ffff) == 0) &&
		((up_speed & 0x0000ffff) == 0)) {

		     down_speed>>=16;
		     up_speed>>=16;

	        } 

		report(0, REPORT_INFO, "ADSL line is up (%d kbit/s down | %d kbit/s up)\n",down_speed,up_speed);
		break;
	}
	
	return(0);

}

/*
* Function      : handle_endpoint_int
* Return Values : None
* Description   : This function handles the interrupt endpoint in a child
*                 process  and returns immediately.
*
*                 6 bytes are received when the line goes UP : a1 00 01 00 00 00
*                 6 bytes are received for DOWN              : a1 00 00 00 00 00
*/
static void handle_endpoint_int(pusb_device_t fdusb)
{

	pid_t child_pid;

	fflush(stdout);
	fflush(stderr);

	if ((child_pid=fork ()) == 0) {

		int up;
		unsigned char buf[TOTAL+1000], prevbuf[TOTAL];
		pusb_endpoint_t ep_int;
		pid_t parent;

		parent = getppid();

		ep_int = pusb_endpoint_open(fdusb, EP_INT, O_RDONLY);
		
		if (ep_int == NULL) {
			report(0, REPORT_ERROR, "pusb_endpoint_open EP_INT\n");
			exit(-1);
		}

#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__) 
		/*
		 * BUG: unless we close the control endpoint, pppoa will
		 *      not be able to open the device.
		 */
		pusb_close(fdusb);
#endif

		up = 0;
		memset(buf, 0, TOTAL + 1000);
		memset(prevbuf, 0, TOTAL);

		while(1) {

			int ret;
			unsigned char lbuf[6];

			/* We can try reading the int endpoint */
			ret = pusb_endpoint_read(ep_int, lbuf, sizeof(lbuf), 0);
			
			/* Ok we failed, perhaps the device has been disconnected  */
			if(ret < 0 && errno == ENODEV) {
				report(0, REPORT_INFO, "Device disconnected, shutting down");
				break;
			}

			/* Just a failure -- report the error and then wait in the waiting loop */
			if(ret < 0)
				report(0, REPORT_ERROR, "Error reading interrupts\n");

			/*
			 * Perhaps the reading is a success, in this case the buffer is 6 bytes
			 * long and the content of the buffer is the line state as described
			 * in the two arrays up_int and down_int.
			 */
			if(ret == 6) {

				/* The magic interrupt for "up state" */
				char up_int[6] = { 0xa1, 0x00, 0x01, 0x00, 0x00, 0x00};
				/* The magic interrupt for "down state" */
				char down_int[6] = { 0xa1, 0x00, 0x00, 0x00, 0x00, 0x00};

				if(!memcmp(up_int, lbuf, 6)) {
					/* We only signal the parent process the first time */
					if(!up) {
						report(1, REPORT_INFO, "Sending a signal to notify the adsl up state\n");
						kill(parent, SIGUSR2);
						up = 1;
					}
					if(signal_kernel) {
						sleep(1);
						pusb_ioctl(fdusb, 1, 1, NULL);
					}
				}

				if(!memcmp(down_int, lbuf, 6)) {
					report(1, REPORT_INFO, "ADSL link went down\n");
					if(signal_kernel) {
						sleep(1);
						pusb_ioctl(fdusb, 1, 2, NULL);
					}
				}
	
			} else {
			
				/*
				 * This method was used in the good old days when benoit was still the
				 * maintainer. I'm putting it back because there are still people
				 * experimenting hangs reading the int endpoint.
				 */

				/* Get the state from the device */
				get_state(fdusb, buf);
			
				if(memcmp(buf, prevbuf, TOTAL)) {

					/* Ok the line is UP */
					if(buf[OFFSET_7] == 0x20 && prevbuf[OFFSET_7] != 0x20) {
						if(!up) {
							report(1, REPORT_INFO, "Sending a signal to notify the adsl up state\n");
							kill(parent, SIGUSR2);
							up = 1;
						}
						if(signal_kernel) {
							sleep(1);
							pusb_ioctl(fdusb, 1, 1, NULL);
						}
					}
				
					/* Bad the line is down */
					if(buf[OFFSET_7] != 0x20 && prevbuf[OFFSET_7] == 0x20) {
						report(1, REPORT_INFO, "ADSL link went down\n");
						if(signal_kernel) {						
							sleep(1);
							pusb_ioctl(fdusb, 1, 2, NULL);
						}
					}
				
					/* Copy new buffer into the previous one */
					memcpy(prevbuf, buf, TOTAL);
				}
			}
		}

		pusb_endpoint_close(ep_int);

		exit(-1);

	}

}

static void signal_usr2(int signal)
{

	link_up = ~link_up;

}

/*****************************************************************************
*	Logging
*****************************************************************************/

void report( int minlevel, unsigned int flags, char *format, ...)
{


	va_list ap;
	int length = 0;
	char *buffer = NULL;


	if( verbose >= minlevel ) {

		/* if log is null, try to open the log file */
		if(flog == NULL) {

#ifndef USE_SYSLOG
			flog = stderr;

			/* set raw mode */
			setbuf(flog, NULL);
#else
			openlog("modem_run", LOG_PID, LOG_USER);
			flog = (FILE*)0xcafecafe;
#endif

		}

		va_start(ap, format);

#ifndef USE_SYSLOG
		/* prints date */
		if(flags & REPORT_DATE) {
			time_t tps;
			time(&tps);
			fprintf(flog, "[%.24s] ", ctime(&tps));
		}


		if(flags&REPORT_ERROR)
			fprintf(flog, "Error ");
		else if (flags&REPORT_INFO)
			fprintf(flog, "Info  ");
		else if (flags&REPORT_DEBUG)
			fprintf(flog, "Debug ");

#endif

		/* Prints log message */
		if(!(flags&REPORT_DUMP)) {
#ifndef USE_SYSLOG
			vfprintf(flog, format, ap);
#else
			char msg[256];

			vsnprintf(msg, 256, format, ap);
			syslog(LOG_INFO, msg);
#endif
		}
		else {

#ifdef USE_SYSLOG
			unsigned char msg[256];
#endif

			buffer = va_arg(ap, char*);
			length = va_arg(ap, int);

#ifndef USE_SYSLOG
			vfprintf(flog, format, ap);
#else
			vsnprintf(msg, 256, format, ap);
			if(flags&REPORT_ERROR)
				syslog(LOG_ERR, msg);
			else if (flags&REPORT_INFO)
				syslog(LOG_INFO, msg);
			else if (flags&REPORT_DEBUG)
				syslog(LOG_DEBUG, msg);
#endif

		}

#ifndef USE_SYSLOG
		/* If needed we include the strerror from libc */
		if(flags & REPORT_PERROR)
			fprintf(flog, "Reason : %s\n", strerror(errno));
#endif

		if(flags & REPORT_DUMP)
			dump(buffer, length);

		va_end(ap);

	}

	return;

}

static void report_stop()
{

#ifdef USE_SYSLOG
	closelog();
#endif

	flog = NULL;

}

/******************************************************************************
* (Sub) Utility functions
******************************************************************************/

static unsigned char give_char(unsigned char c)
{

	if(c >= ' ' && c < 0x7f)
		return(c);
	else
		return('.');

}

static void dump(unsigned char *buf, int len)
{

	int i, j;

#ifndef USE_SYSLOG
	if(flog == NULL)
		return;
#else
	unsigned char line[128];
	unsigned char *ptr;
#endif

	if(buf == NULL || len == 0)
		return;

	for(i = 0; i < len; i += 16) {

#ifdef USE_SYSLOG
		ptr = line;
#endif

		for(j = i; j < len && j < i + 16; j++) {
#ifndef USE_SYSLOG
			fprintf(flog, "%02x ", buf[j]);
#else
			sprintf(ptr, "%02x ", buf[j]);
			ptr += 3;
#endif
		}

		for(; j < i + 16; j++) {
#ifndef USE_SYSLOG
			fprintf(flog, "   ");
#else
			sprintf(ptr, "   ");
			ptr += 3;
#endif
		}

		for(j = i; j < len && j < i + 16; j++) {
#ifndef USE_SYSLOG
			fprintf(flog, "%c", give_char(buf[j]));
#else
			sprintf(ptr, "%c", give_char(buf[j]));
			ptr++;
#endif
		}

#ifndef USE_SYSLOG
		fprintf(flog, "\n");
#else
		*ptr = '\0';
		syslog(LOG_DEBUG, line);
#endif

	}
	
#ifndef USE_SYSLOG
	fprintf(flog, "\n");
#endif

}

#endif /* _MODEM_RUN_C_*/
