/*
*  ALCATEL SpeedTouch USB modem utility : PPPoA implementation (2nd edition)
*  Copyright (C) 2001 Benoit PAPILLAULT
*  
*  This program is free software; you can redistribute it and/or
*  modify it under the terms of the GNU General Public License
*  as published by the Free Software Foundation; either version 2
*  of the License, or (at your option) any later version.
*  
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
*  Author : Benoit PAPILLAULT <benoit.papillault@free.fr>
*  Creation : 21/03/2001
*
*  This program is designed to work under pppd, with the option "pty".
*
*  $Id: pppoa2.c,v 1.19 2002/12/05 21:47:16 edgomez Exp $
*/

#ifndef _PPPOA2_C_
#define _PPPOA2_C_

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <dirent.h>
#include <termios.h>		/* N_HDLC & TIOCSETD */
#include <string.h>
#include <limits.h>		/* for LONG_MAX */
#include <stdarg.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/resource.h>	/* setpriority() */

#ifdef USE_SYSLOG
#include <syslog.h>
#endif


/* Portable USB library */
#include "pusb.h"

/* ATM lib */
#include "atm.h"

/* Small semaphore library */
#include "smallsem.h"

/* For report prototype */
#include "pppoa3.h"

/*****************************************************************************
*	Defines
*****************************************************************************/

/* Key of the log's mutex */
#define	LOG_SEMKEY 0xCAFEDECA

/* USB timeout */
#define DATA_TIMEOUT 1000

/* Buffer size of buffer */
#define BUFFER_SIZE (64 * 1024)
#define HDLC_HEADER_SIZE 2

/* PPP level : DO NOT MODIFY */
#define syncHDLC 1

/*****************************************************************************
*	Variables
*****************************************************************************/

/* ATM Virtual Path Id and Virtual Circuit Id */
static int my_vpi = -1;
static int my_vci = -1;

/* Exclusive acces when outputing to stdout */
static int log_mutex=0;

/* Read and write buffers */
static unsigned char ppp_read_buf[HDLC_HEADER_SIZE + 1367*ATM_CELL_TOTAL_SIZE];
static unsigned char ppp_write_buf[HDLC_HEADER_SIZE + 1367*ATM_CELL_TOTAL_SIZE];
static unsigned char *aal5_send_buf = &ppp_read_buf[HDLC_HEADER_SIZE];
static unsigned char *aal5_recv_buf = &ppp_write_buf[HDLC_HEADER_SIZE];

/* Global parameters */
static int   verbose = 0;

/* processes pid_ts */
static pid_t this_process;		/*always the current   process pid*/
static pid_t parent_process;		/*always the parent    process pid*/
static pid_t handler87_process;		/*always the handler87 process pid*/

/* TODO : remove this var */
static int gfdout;
static int alternate_ep = -1;

/*****************************************************************************
*	Prototypes
*****************************************************************************/

static unsigned char give_char(unsigned char c);

static int ppp_read(int fd, unsigned char *buffer, int n);
static int ppp_write(int fd, unsigned char *buffer, int n);

static void usage();
static void sighandler(int signal);
static void dump(unsigned char *buf, int len);
#if 0
static void handle_endpoint_81(pusb_endpoint_t ep_int);
#endif
static void handle_endpoint_87(pusb_endpoint_t epdata, int fdout);
static void handle_endpoint_07(pusb_endpoint_t epdata, int fdin);

/*****************************************************************************
*       Main function
*****************************************************************************/

/*
* Function      : main
* Return Values : returns only if error is detected (error codes are != 0)
* Description   : Initializes the HDLC, log, and handle endpoints 07, 87
*                 then exit the main process.
*/
int main(int argc, char *argv[])
{
	int    fdin, fdout;
	int    i;

#ifndef USE_SYSLOG
	time_t ourtime;
	char   *logfile;
	int log;
#endif

	pusb_device_t fdusb;
	pusb_endpoint_t ep_data;
	char   *devicename = NULL;

#ifndef USE_SYSLOG
	logfile = "/var/log/pppoa2.log";
	log = 0;
#endif

	for(i = 1; i < argc; i++)
	{
		if(strcmp(argv[i], "-v") == 0 && i + 1 < argc)
			verbose = atoi(argv[++i]);
		else if(strcmp(argv[i], "-vpi") == 0 && i + 1 < argc)
			my_vpi = atoi(argv[++i]);
		else if(strcmp(argv[i], "-vci") == 0 && i + 1 < argc)
			my_vci = atoi(argv[++i]);
		else if(strcmp(argv[i], "-d") == 0 && i + 1 < argc)
			devicename = argv[++i];
		else if(strcmp(argv[i], "-e") == 0 && i + 1 < argc)
			alternate_ep = atoi(argv[++i]);
		else if(strcmp(argv[i], "--help") == 0)
			usage();
#ifndef USE_SYSLOG
		else if(strcmp(argv[i], "-f") == 0 && i + 1 < argc)
			logfile   = argv[++i];
#endif
		else 
			usage();
	}

	/*vpi & vci aren't set*/
	if(my_vpi == -1 || my_vci == -1)
		usage();

	/*Duplicate in and out fd*/
	fdin  = dup(STDIN_FILENO);
	fdout = dup(STDOUT_FILENO);

	/*Get process pid*/
	parent_process = this_process = getpid();

#ifndef USE_SYSLOG
	/* Create the log */
	log = open(logfile, O_CREAT | O_RDWR | O_APPEND, 0600);

	/* Failed to create the log file*/
	if(log < 0)
	{
		perror(logfile);
		return(-1);
	}
#else
	openlog("pppoa2", LOG_PID, LOG_DAEMON);
#endif

	/*
	 * Check if our mutex remains from previous executions
	 * of pppoa2. If yes, destroy it. 
	 * Then create a fresh one and set its initial value to 1.
	 */
	if ((log_mutex=sem_get((key_t)LOG_SEMKEY))!=-1)
	{
		if (sem_destroy(log_mutex)==-1)
		{
#ifndef USE_SYSLOG
			perror("sem_destroy");
#else
			syslog(LOG_ERR,"sem_destroy");
#endif
			return(-1);
		}
	}
	if ((log_mutex=sem_create((key_t)LOG_SEMKEY))==-1)
	{
#ifndef USE_SYSLOG
		perror("sem_create");
#else
		syslog(LOG_ERR, "sem_create");
#endif
		return(-1);
	}
	if (sem_init(log_mutex, 1)==-1)
	{
#ifndef USE_SYSLOG
		perror("sem_init");
#else
		syslog(LOG_ERR, "sem_init");
#endif
		return(-1);
	}

	/*
	* we will redirect all standard streams to the log file
	* so we are sure to catch all messages from this program :)
	*/
#ifndef USE_SYSLOG
	if(dup2(log, STDIN_FILENO) == -1)
	{
		write(log,"Error duplicating STDIN_FILENO\n", 31);
		return(-1);
	}
	if(dup2(log, STDOUT_FILENO) == -1)
	{
		write(log,"Error duplicating STDOUT_FILENO\n", 32);
		return(-1);
	}
	if(dup2(log, STDERR_FILENO) == -1)
	{
		write(log,"Error duplicating STDERR_FILENO\n", 32);
		return(-1);
	}

	/*We don't need log fd anymore*/
	if(close(log) == -1)
	{
		fprintf(stderr, "Error closing log fd\n");
		return(-1);
	}

	/*
	* No buffering on stdout & stderr
	* So the stdout and stderr are flushed
	* each time we write to them
	*/
	setbuf(stdout, NULL);
	setbuf(stderr, NULL);

#endif

	report(0, REPORT_INFO, "Starting PPPoA2 ( merged version includes new ATM/AAL5 stack ) %s\n", VERSION);

#ifndef USE_SYSLOG
	time(&ourtime);
	report(0, REPORT_INFO, "Log started on %s\n", ctime(&ourtime));
#endif

	/* If level > 1, write the vpi,vci used */
	report(2, REPORT_INFO, "Using vpi=%d, vci=%d\n", my_vpi, my_vci);

	/* To know what pid is the parent process */
	report(0, REPORT_INFO, "I'm the parent   process, I handle the endpoint 0x07\n", parent_process);

	/* Give information about the tty fds */
	report(1, REPORT_DEBUG, "pty descriptors : fdin=%d, fdout=%d\n", fdin, fdout);

	/* Increase priority of the pppoa process*/
	if(setpriority(PRIO_PROCESS, this_process, -20) < 0)
		report(1, REPORT_INFO|REPORT_DATE|REPORT_PERROR,"setpriority failed\n");

#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
#define SOCKBUF (64*1024)
	{
		int sbuf, ss = sizeof(sbuf);

		if(getsockopt(fdin, SOL_SOCKET, SO_SNDBUF, &sbuf, &ss) == 0)
		{

			/*Debug message*/
			report(1, REPORT_INFO|REPORT_DATE, "Increasing SNDBUF from %d to %d\n", sbuf, SOCKBUF);

			sbuf = SOCKBUF;

			if(setsockopt(fdin, SOL_SOCKET, SO_SNDBUF, &sbuf, ss) < 0)
				report(0, REPORT_ERROR|REPORT_DATE|REPORT_PERROR, "setsockopt failed\n");

		}

		if(getsockopt(fdin, SOL_SOCKET, SO_RCVBUF, &sbuf, &ss) == 0)
		{

			/*Debug message*/
			report(1, REPORT_INFO|REPORT_DATE, "Increasing RCVBUF from %d to %d\n", sbuf, SOCKBUF);

			sbuf = SOCKBUF;

			if(setsockopt(fdin, SOL_SOCKET, SO_RCVBUF, &sbuf, ss) < 0)
				report(0, REPORT_ERROR|REPORT_DATE|REPORT_PERROR, "setsockopt failed\n");
		}
	}
#endif

	/*
	* Install HDLC line discipline on fdin if it is a tty and
	* the OS has such a thing.
	*/
	if(isatty(fdin))
	{
#ifdef N_HDLC
		int disc = N_HDLC;
		if(ioctl(fdin, TIOCSETD, &disc) < 0)
		{
			report(0, REPORT_ERROR|REPORT_DATE, "Error loading N_HDLC\n");
			return(-1);
		}
		report(2, REPORT_INFO|REPORT_DATE, "N_HDLC line set up\n");

#elif defined TTYDISC
		int disc = TTYDISC;
		if(ioctl(fdin, TIOCSETD, &disc) < 0)
		{
			report(0, REPORT_ERROR|REPORT_DATE, "Error setting termios tty line discipline\n");
			return(-1);
		}
		report(2, REPORT_INFO|REPORT_DATE, "TTYDISC line set up\n");

#endif
	}

	if (devicename == NULL) {
	/*
	*  We search for the first USB device matching ST_VENDOR & ST_PRODUCT.
	*  usbdevfs must be mount on /proc/bus/usb (or you may change the path
	*  here, according to your config
	*/

		fdusb = pusb_search_open(ST_VENDOR, ST_PRODUCT);

		if(fdusb == NULL)
		{
			report(0, REPORT_ERROR|REPORT_DATE, "Where is this crappy modem ?!\n");
			return(-1);
		}

	} else {

		fdusb = pusb_open(devicename);

		if(fdusb == NULL) {
			report(0, REPORT_ERROR|REPORT_DATE, "The modem is not at %s\n", devicename);
			return(-1);
		}

	}

	/*Debug message*/
	report(2, REPORT_DEBUG|REPORT_DATE, "Got the modem !\n");

	/* Initialize global variables */
	gfdout = fdout;

	/* We claim interface 1, where endpoints 0x07 & 0x87 are */
	if(pusb_claim_interface(fdusb, 1) < 0)
	{
		report(0, REPORT_ERROR|REPORT_DATE, "pusb_claim_interface 1 failed\n");
		return(-1);
	}

	if (alternate_ep != -1) {
	    /* Historically this wasn't set */
	    /* reconfigure USB (configuration & alternate settings) */
	    if (pusb_set_configuration(fdusb,1) < 0) {
		    report(0, REPORT_ERROR, "pusb_set_configuration 1");
		    return(-1);
	    }
	    
	    if (pusb_set_interface(fdusb,1,alternate_ep) < 0) {
		    report(0, REPORT_ERROR, "pusb_set_interface");
		    return(-1);
	    }
	}

	ep_data = pusb_endpoint_open(fdusb, EP_DATA_OUT, O_RDWR);
	if(ep_data == NULL)
	{
		report(0, REPORT_ERROR|REPORT_DATE, "pusb_endpoint_open failed\n");
		return(-1);
	}

	/*
	* endpoint 0x87 is used for receiving ATM cells, this function will
	* fork a new process that will read ATM cells from the usb and send them
	* to pppd
        */

	handle_endpoint_87(ep_data, fdout);

	/* At least in FreeBSD, ppp may kill us with a hangup. */
	/* In Linux, pppd kill us with a term signal ! */
#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
	signal(SIGHUP , sighandler);
#elif defined(__linux__)
	signal(SIGTERM, sighandler);
#endif

	/*
	* endpoint 0x07 is used for sending ATM cells, this function
	* is an infinite loop and will return only on errors.
        */

	handle_endpoint_07(ep_data, fdin);

	/*
	*  we kill our child pid : this is to be sure that
	*  the USB device file will be available for other programs
	*/

	if(kill(handler87_process, SIGINT) < 0)
		report(0, REPORT_ERROR|REPORT_DATE|REPORT_PERROR, "Error killing child process %d\n", handler87_process);

	/* we release all the interface we'd claim before exiting */
	if(pusb_release_interface(fdusb, 1) < 0)
		report(0, REPORT_ERROR|REPORT_DATE|REPORT_PERROR,"pusb_release_interface failed\n");

	pusb_endpoint_close(ep_data);

	pusb_close(fdusb);

	sem_destroy(log_mutex);	/* destroys the semaphore from the system */

	return(0);

}

/*****************************************************************************
*	Logging functions
*****************************************************************************/

static unsigned char give_char(unsigned char c)
{

	if(c >= ' ' && c < 0x7f)
		return(c);
	else
		return('.');

}

/*
* Function      : dump
* Return Values : none
* Description   : Dump the buffer in a log file / syslog
*/
static void dump(unsigned char *buf, int len)
{

	int i, j;

#ifdef USE_SYSLOG
	unsigned char line[128];
	unsigned char *ptr;
#endif

	if(buf == NULL || len == 0)
		return;

	for(i = 0; i < len; i += 16) {

#ifdef USE_SYSLOG
		ptr = line;
#endif

		for(j = i; j < len && j < i + 16; j++) {
#ifndef USE_SYSLOG
			printf("%02x ", buf[j]);
#else
			sprintf(ptr, "%02x ", buf[j]);
			ptr += 3;
#endif
		}

		for(; j < i + 16; j++) {
#ifndef USE_SYSLOG
			printf("   ");
#else
			sprintf(ptr, "   ");
			ptr += 3;
#endif
		}

		for(j = i; j < len && j < i + 16; j++) {
#ifndef USE_SYSLOG
			printf("%c", give_char(buf[j]));
#else
			sprintf(ptr, "%c", give_char(buf[j]));
			ptr++;
#endif
		}

#ifndef USE_SYSLOG
		printf("\n");
#else
		*ptr = '\0';
		syslog(LOG_DEBUG, line);
#endif

	}
	
#ifndef USE_SYSLOG
	printf("\n");
#endif

}

/*
* Function      : report
* Return Values : none
* Description   : Logs a message if verbose level is higher than
*                 the minlevel required
*/
void report( int minlevel, unsigned int flags, char *format, ...)
{
	va_list ap;
	int length = 0;
	char *buffer = NULL;

	if( verbose >= minlevel ) {

		/* lock the report mutex to avoid log collisions */
		sem_P(log_mutex);

		va_start(ap, format);

#ifndef USE_SYSLOG
		/* prints date */
		if(flags & REPORT_DATE) {
			time_t tps;
			time(&tps);
			printf("[%.24s] ", ctime(&tps));
		}

		if(flags&REPORT_ERROR)
			printf("Error ");
		else if (flags&REPORT_INFO)
			printf("Info  ");
		else if (flags&REPORT_DEBUG)
			printf("Debug ");

		if(flags)
			printf(">%d< ", (int)this_process); 

#endif
		/* Prints log message */
		if(!(flags&REPORT_DUMP)) {
#ifndef USE_SYSLOG
			vprintf(format, ap);
#else
			char msg[256];

			vsnprintf(msg, 256, format, ap);
			syslog(LOG_INFO, msg);
#endif
		}
		else {
#ifdef USE_SYSLOG
			unsigned char msg[256];
#endif
			buffer = va_arg(ap, char*);
			length = va_arg(ap, int);

#ifndef USE_SYSLOG
			vprintf(format, ap);
#else
			vsnprintf(msg, 256, format, ap);
			if(flags&REPORT_ERROR)
				syslog(LOG_ERR, msg);
			else if (flags&REPORT_INFO)
				syslog(LOG_INFO, msg);
			else if (flags&REPORT_DEBUG)
				syslog(LOG_DEBUG, msg);
#endif
		}

#ifndef USE_SYSLOG
		/* If needed we include the strerror from libc */
		if(flags & REPORT_PERROR)
			printf("Reason : %s\n", strerror(errno));
#endif

		if(flags & REPORT_DUMP)
			dump(buffer, length);

		va_end(ap);

		/* UnLock the report mutex */
		sem_V(log_mutex);
	}

	return;

}

/******************************************************************************
*	PPP i/o functions
******************************************************************************/

/*
* variable that will be used in several function
*   => so global
*/


/*
* Function     : ppp_read
* Return value : data read length
* Description  : Reads size bytes from fd and put them in the buffer
*                It sync hdlc 
*/

static int ppp_read(int fd, unsigned char *buffer, int size)
{
	/*
	* we must handle HDLC framing, since this is what pppd
	* sends to us. I use some code from rp-pppoe
	*/

	int r;

	if(size > BUFFER_SIZE) {
		report(0, REPORT_ERROR|REPORT_DATE, "Error %d bytes requested but the buffer is %d bytes long\n", size, BUFFER_SIZE);
		return(0);
	}
	for(;;) {
		do {
			r = read(fd, buffer, size);
		} while(r < 0 && errno == EINTR);

		if(r < 0) {
			report(0, REPORT_ERROR|REPORT_DATE|REPORT_PERROR, "Error reading from pppd\n");
			return(-1);
		}

		if(!r)
			return(0);

		/* supress leading 2 bytes */
		if(syncHDLC) {

			if(r < 3) {
				report(0, REPORT_ERROR|REPORT_DATE, "Read from ppp short on data, r=%d => ignored\n", r);
				continue;
			}

			if(r == BUFFER_SIZE) {
				report(0, REPORT_ERROR|REPORT_DATE, "Read from ppp too much data, r=%d => ignored\n", r);
				continue;
			}

			return(r - HDLC_HEADER_SIZE);
		}
		else {
			report(0, REPORT_ERROR|REPORT_DATE, "async HDLC not handled\n");
		}
		break;
	}
}

/*
* Function     : ppp_write
* Return value : data written length
* Description  : Writes size bytes to fd
*/
/*
* Define the number of time we try to send a packet when there isn't
* any buffer left
*/
#define NOBUF_RETRIES 5
static int ppp_write(int fd, unsigned char *buffer, int n)
{
	int r;
	static int errs = 0;
	int retries = 0;

	if(syncHDLC) {
		retry:
		if((r = write(fd, buffer, n)) < 0) {

			/*
			* We sometimes run out of mbufs doing simultaneous
			* up- and down-loads on FreeBSD.  We should find out
			* why this is, but for now...
			*/
			if(errno == ENOBUFS) {
				errs++;
				if((errs < 10 || errs % 100 == 0))
					report(0, REPORT_ERROR|REPORT_DATE, "ppp_write: %d ENOBUFS errors\n", errs);

				if(retries++ < NOBUF_RETRIES) {
					/* retry after delay */
					usleep(1000);
					goto retry;
				}
				else {
					/* throw away the packet */
					return(0);
				}
			}
			else {
				report(0, REPORT_ERROR|REPORT_DATE|REPORT_PERROR, "Error writing to ppp(d)\n");
			}
		}
		return(r);
	}
	else {
		report(0, REPORT_ERROR|REPORT_DATE, "async HDLC not handled\n");
	}
	return(-1);
}

/******************************************************************************
*	Endpoints functions
******************************************************************************/

/*
* Function      : handle_endpoint_81
* Return Values : none
* Description   : endpoint 0x81, which is on interface 0 is used for 
*                 handling 'interrupt'.
* Usage         : Currently we don't use that function as it's handle in
*                 Modem_run
*/
#if 0
static void handle_endpoint_81(pusb_endpoint_t ep_int)
{
	pid_t child_pid;

	if((child_pid = fork()) == 0)
	{
		unsigned char lbuf[64 * 53];
		int ret;

		for(;;)
		{
			ret = pusb_endpoint_read(ep_int, lbuf, sizeof(lbuf), 0);

			if(ret < 0)
			{
				report(0, REPORT_ERROR|REPORT_DATE, "Error reading interrupts\n");
				break;
			}

			report(2, REPORT_DEBUG|REPORT_DATE|REPORT_DUMP, "received interrupts, len = %d\n", lbuf, ret, ret);
		}
		_exit(0);
	}
}
#endif

/*
* Function      : handle_endpoint_87
* Return Values : none
* Description   : to handle synchronous read, we fork a new process
*                 which does an infinite loop, keeping reading on
*                 endpoint EP_DATA_IN.
*/
static void handle_endpoint_87(pusb_endpoint_t epdata, int fdout)
{
	pid_t child_pid;

	if((child_pid = fork()) == 0)
	{
		int pos;

		/*
		* This is the children part of the fork()
		* child_pid doesn't mean nothing, just we are a child process
		* so we get our real pid
		*/
		this_process = handler87_process = getpid();

		report(0, REPORT_INFO, "I'm the children process, I handle the endpoint 0x87\n", this_process);
		
		/* Prepares ppp buffer for HDLC */
		ppp_write_buf[0] = 0xff;
		ppp_write_buf[1] = 0x03;

		/* Prepares pos */
		pos = 0;

		/* Reads from usb and writes to ppp(d) tty */
		for(;;) {
			int n;
			int pti;
			unsigned char lbuf[64*53];
			unsigned char *unused_cells;

			/* Reads 64*53 bytes from usb */
			do {
				n = pusb_endpoint_read(epdata, lbuf, sizeof(lbuf), 0);
			} while (n < 0 && (errno == EINTR ||errno == ETIMEDOUT));

			if(n < 0) {
				report(0, REPORT_ERROR|REPORT_DATE|REPORT_PERROR, "Error reading usb urb\n");
				break;
			}

			/* Debug information */
			report(2, REPORT_DEBUG|REPORT_DATE|REPORT_DUMP, "ATM cells read from USB (%d bytes long)\n", lbuf, n, n);

			/* Accumulates data in the aal5_recv buffer */
			/* pti will be  equal to the last cell pti */
			pti = aal5_frame_from_atm_cells(aal5_recv_buf, lbuf, n, my_vpi, my_vci, &pos, &unused_cells);
	
			/* A buffer overflow has been detected */
			if(pti<0) {
				report(0, REPORT_ERROR|REPORT_DATE, "Buffer overflow, too many cells for the same aal5 frame\n");
			}

			/* As the last pti is 1, we have to send the aal5_frame data */
			while(pti) {

				/* Debug information */
				report(2, REPORT_DEBUG|REPORT_DATE|REPORT_DUMP, "AAL5 frame joined up  (%d bytes long)\n", aal5_recv_buf, pos, pos);

				/* Prepares the aal5 data (no overwrite is done)*/
				n = aal5_frame_dec(aal5_recv_buf, aal5_recv_buf, pos);

				if(n<0) {

					report(0, REPORT_ERROR|REPORT_DATE, "CRC error in an AAL5 frame\n");
				}
				else {
					/* Writes the result buffer */
					if(ppp_write(gfdout, ppp_write_buf, n + HDLC_HEADER_SIZE) > 0)
						report(2, REPORT_DEBUG|REPORT_DATE, "Extracted PPP packet sent to ppp(d)\n\n");
				}
				/* Reinitializes pos and n*/
				pos = 0;
				pti = 0;

				/* Are cells unused in lbuf ? */
				if(unused_cells != NULL) {
					int length;

					length = sizeof(lbuf) - ((int)unused_cells - (int)lbuf);
					pti = aal5_frame_from_atm_cells(aal5_recv_buf, unused_cells, length, my_vpi, my_vci, &pos, &unused_cells);
				}
			}
		}

		/*
		* this line is reached only if an error occurs in
		* communicating with usb or pppd
		*/

		/* first, we kill our father, which is useless */
		if(kill(parent_process, SIGINT) < 0)
			report(0, REPORT_DATE|REPORT_PERROR, "Error killing parent process\n");

		sem_destroy(log_mutex);	/* destroys the semaphore from the system */

		/* now we exit from this process too*/
		_exit(0);

	}
	else {

		/*
		* This is the parent part of the fork
		* child_pid is the pid of the forked process
		* we save it in handler87_process to be able to kill it
		* when a signal is handled by sighandler()
		*/

		handler87_process = child_pid;

	}

}

/*
* Function      : handle_endpoint_07
* Return Values : none
* Description   : to handle synchronous write, we does an infinite
*                 loop, keeping writing on endpoint EP_DATA_OUT
* Usage         : this function must be called after the
*                 handle_endpoint_87 which is a fork.
*                 handle_endpoint_07 is an infinite loop
*                 so if you enter it, you will never get
*                 out ( just when there are errors)
*/
static void handle_endpoint_07(pusb_endpoint_t epdata, int fdin)
{
	/* this part still need a redesign ... */
	int n;

	for(;;) {
		/* Reads data from ppp(d) */
		n = ppp_read(fdin, ppp_read_buf, BUFFER_SIZE);

		if(n <= 0)
			break;

		/* Debug information */
		report(2, REPORT_DEBUG|REPORT_DATE|REPORT_DUMP, "PPP packet read from ppp(d) (%d bytes long)\n", aal5_send_buf, n, n);

		/* Creates the aal5 frame */
		n = aal5_frame_enc(aal5_send_buf, aal5_send_buf, n);

		/* Builds a queue of atm cells from the aal5 frame */
		n = aal5_frame_to_atm_cells(aal5_send_buf, aal5_send_buf, n, my_vpi, my_vci);

		if(n < 0) {
			report(0 , REPORT_ERROR|REPORT_DATE, "Error pppoa2 is buggy\n");
			break;
		}

		/* Debug information */
		report(2, REPORT_DEBUG|REPORT_DATE|REPORT_DUMP, "ATM cell queue built (%d bytes long)\n", aal5_send_buf, n, n);

		/* Sends data on the usb bus */
		n = pusb_endpoint_write(epdata, aal5_send_buf, n, DATA_TIMEOUT);

		if(n > 0)
			report(2, REPORT_DEBUG|REPORT_DATE, "ATM cell queue sent to USB\n\n");
	}
	return;
}

/*
* Function      : usage
* Return Values : None
* Description   : Prints an usage message
*/
static void usage()
{
	fprintf(stderr, "pppoa2 version %s\n", VERSION);
#ifndef USE_SYSLOG
	fprintf(stderr, "usage: pppoa2 [-v level] [-f logfile] -vpi val -vci val\n");
#else
	fprintf(stderr, "usage: pppoa2 [-v level] -vpi val -vci val\n");
#endif
	fprintf(stderr, "  -v       : defines the verbosity level [0-2]\n");
#ifndef USE_SYSLOG
	fprintf(stderr, "  -f       : defines the log filename to use ( Default /tmp/pppoa2.log )\n");
#endif
	fprintf(stderr, "  -vpi.vci : defined the vpi.vci that your provider is\n");
	fprintf(stderr, "             using, 8.35 for France, 0.38 for UK\n");
	fprintf(stderr, "  -d device: defines the device to use\n");
	fprintf(stderr, "  -e number: use an alternate endpoint interface (1 for a SpeedTouch USB 330)\n");
	fprintf(stderr, "  --help   : this message\n");
	exit(-1);
}

/*
* Function      : sighandler
* Return Values : None
* Description   : Die on a hangup, and propagate it to the handler_87 process
*/
static void sighandler(int signal)
{
	/*We kill the handler87 handler process*/
	if(kill(handler87_process, signal) < 0)
		report(0, REPORT_ERROR|REPORT_DATE|REPORT_PERROR, "Error killing handler87 process\n");

	report(0, REPORT_INFO|REPORT_DATE, "Killed by a signal %d\n", signal);

	sem_destroy(log_mutex);	/* destroys the semaphore from the system */

	/* We exit safely*/
	_exit(1);
}

#endif /* _PPPOA2_C_ */
